﻿using System;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Utilities;

namespace Org.BouncyCastle.Tls.Crypto.Impl.BC
{
    /// <summary>BC light-weight support class for handling TLS secrets and deriving key material and other secrets
    /// from them.</summary>
    public class BcTlsSecret
        : AbstractTlsSecret
    {
        public static BcTlsSecret Convert(BcTlsCrypto crypto, TlsSecret secret)
        {
            if (secret is BcTlsSecret)
                return (BcTlsSecret)secret;

            if (secret is AbstractTlsSecret)
            {
                AbstractTlsSecret abstractTlsSecret = (AbstractTlsSecret)secret;

                return crypto.AdoptLocalSecret(CopyData(abstractTlsSecret));
            }

            throw new ArgumentException("unrecognized TlsSecret - cannot copy data: " + Core.Platform.GetTypeName(secret));
        }

        protected readonly BcTlsCrypto m_crypto;

        public BcTlsSecret(BcTlsCrypto crypto, byte[] data)
            : base(data)
        {
            this.m_crypto = crypto;
        }

        public override TlsSecret DeriveUsingPrf(int prfAlgorithm, string label, byte[] seed, int length)
        {
            lock (this)
            {
                CheckAlive();

                switch (prfAlgorithm)
                {
                case PrfAlgorithm.tls13_hkdf_sha256:
                    return TlsCryptoUtilities.HkdfExpandLabel(this, CryptoHashAlgorithm.sha256, label, seed, length);
                case PrfAlgorithm.tls13_hkdf_sha384:
                    return TlsCryptoUtilities.HkdfExpandLabel(this, CryptoHashAlgorithm.sha384, label, seed, length);
                case PrfAlgorithm.tls13_hkdf_sm3:
                    return TlsCryptoUtilities.HkdfExpandLabel(this, CryptoHashAlgorithm.sm3, label, seed, length);
                default:
                    return m_crypto.AdoptLocalSecret(Prf(prfAlgorithm, label, seed, length));
                }
            }
        }

        public override TlsSecret HkdfExpand(int cryptoHashAlgorithm, byte[] info, int length)
        {
            lock (this)
            {
                if (length < 1)
                    return m_crypto.AdoptLocalSecret(TlsUtilities.EmptyBytes);

                int hashLen = TlsCryptoUtilities.GetHashOutputSize(cryptoHashAlgorithm);
                if (length > (255 * hashLen))
                    throw new ArgumentException("must be <= 255 * (output size of 'hashAlgorithm')", "length");

                CheckAlive();

                byte[] prk = m_data;

                FipsShs.AuthenticationParameters parameterSet = m_crypto.GetHmacParameters(cryptoHashAlgorithm);
                FipsShs.Key serviceKey = new FipsShs.Key(parameterSet, prk);
                IStreamCalculator<IBlockResult> hmac = CryptoServicesRegistrar
                    .CreateService(serviceKey)
                    .CreateMacFactory(parameterSet)
                    .CreateCalculator();

                byte[] okm = new byte[length];

                byte[] t = new byte[hashLen];
                byte counter = 0x00;

                int pos = 0;
                for (;;)
                {
                    hmac.Stream.Write(info, 0, info.Length);
                    hmac.Stream.WriteByte(++counter);
                    hmac.Stream.Close();
                    hmac.GetResult().Collect(t, 0);

                    int remaining = length - pos;
                    if (remaining <= hashLen)
                    {
                        Array.Copy(t, 0, okm, pos, remaining);
                        break;
                    }

                    Array.Copy(t, 0, okm, pos, hashLen);
                    pos += hashLen;
                    hmac.Stream.Write(t, 0, t.Length);
                }

                return m_crypto.AdoptLocalSecret(okm);
            }
        }

        public override TlsSecret HkdfExtract(int cryptoHashAlgorithm, TlsSecret ikm)
        {
            lock (this)
            {
                CheckAlive();

                byte[] salt = m_data;
                this.m_data = null;

                FipsShs.AuthenticationParameters parameterSet = m_crypto.GetHmacParameters(cryptoHashAlgorithm);
                FipsShs.Key serviceKey = new FipsShs.Key(parameterSet, salt);
                IStreamCalculator<IBlockResult> hmac = CryptoServicesRegistrar
                    .CreateService(serviceKey)
                    .CreateMacFactory(parameterSet)
                    .CreateCalculator();

                Convert(m_crypto, ikm).UpdateMac(hmac);

                hmac.Stream.Close();
                byte[] prk = hmac.GetResult().Collect();

                return m_crypto.AdoptLocalSecret(prk);
            }
        }

        protected override AbstractTlsCrypto Crypto
        {
            get { return m_crypto; }
        }

        protected virtual byte[] Prf(int prfAlgorithm, string label, byte[] seed, int length)
        {
            if (PrfAlgorithm.ssl_prf_legacy == prfAlgorithm)
                throw new ArgumentException("Unsupported PRF algorithm: " + PrfAlgorithm.GetText(prfAlgorithm));

            if (PrfAlgorithm.tls_prf_legacy == prfAlgorithm)
                return Prf_1_0(label, seed, length);

            return Prf_1_2(prfAlgorithm, label, seed, length);
        }

        protected virtual byte[] Prf_1_0(string label, byte[] seed, int length)
        {
            return CryptoServicesRegistrar.CreateService(FipsKdf.Tls1_0)
                .From(m_data, label, seed)
                .GetResult(length)
                .Collect();
        }

        protected virtual byte[] Prf_1_2(int prfAlgorithm, string label, byte[] seed, int length)
        {
            int cryptoHashAlgorithm = TlsCryptoUtilities.GetHashForPrf(prfAlgorithm);
            var hmacParameters = m_crypto.GetHmacParameters(cryptoHashAlgorithm);

            return CryptoServicesRegistrar.CreateService(FipsKdf.Tls1_2)
                .WithPrf(hmacParameters)
                .From(m_data, label, seed)
                .GetResult(length)
                .Collect();
        }

        protected virtual void UpdateMac(IStreamCalculator<IBlockResult> mac)
        {
            lock (this)
            {
                CheckAlive();

                mac.Stream.Write(m_data, 0, m_data.Length);
            }
        }
    }
}
